import unittest

import numpy as np
import pandas as pd

from ..transform import get_player_totals, merge_with_prev_year


class TestTransform(unittest.TestCase):

    def test_get_player_totals(self):
        mock_df = pd.DataFrame(
            [
                ['harry', 2018, 'kensho', 1],
                ['harry', 2019, 'kensho', 2],
                ['harry', 2019, 'TOT', 3],
                ['harry', 2019, 'ri', 4],
                ['hallas', 2018, 'harvard', 5],
                ['hallas', 2019, 'harvard', 6],
            ],
            columns=['name', 'year', 'team_id', 'var1']
        )

        expected_df = pd.DataFrame(
            [
                ['harry', 2018, 'kensho', 1],
                ['harry', 2019, 'TOT', 3],
                ['hallas', 2018, 'harvard', 5],
                ['hallas', 2019, 'harvard', 6],
            ],
            columns=['name', 'year', 'team_id', 'var1']
        )

        output_df = get_player_totals(mock_df, player_col='name')
        pd.testing.assert_frame_equal(expected_df, output_df)

        with self.assertRaises(ValueError):
            get_player_totals(mock_df)

        # Test with no values of 'TOT'
        mock_df = pd.DataFrame(
            [
                ['harry', 2018, 'kensho', 1],
            ], columns=['name', 'year', 'team_id', 'var1']
        )

        expected_df = pd.DataFrame(
            [
                ['harry', 2018, 'kensho', 1],
            ], columns=['name', 'year', 'team_id', 'var1']
        )
        output_df = get_player_totals(mock_df, player_col='name')
        pd.testing.assert_frame_equal(expected_df, output_df)

    def test_merge_with_prev_year(self):
        mock_df = pd.DataFrame(
            [
                ['harry', 2018, 'kensho', 1],
                ['harry', 2019, 'TOT', 3],
                ['hallas', 2017, 'harvard', 3],
                ['hallas', 2018, 'harvard', 5],
                ['hallas', 2019, 'harvard', 6],
            ],
            columns=['player_url', 'year_col', 'team_id', 'var1']
        )

        expected_df = pd.DataFrame(
            [
                ['harry', 2018, 'kensho', 1, np.nan, np.nan],
                ['harry', 2019, 'TOT', 3, 'kensho', 1],
                ['hallas', 2017, 'harvard', 3, np.nan, np.nan],
                ['hallas', 2018, 'harvard', 5, 'harvard', 3],
                ['hallas', 2019, 'harvard', 6, 'harvard', 5],
            ],
            columns=['player_url', 'year_col', 'team_id', 'var1', 'team_id___1', 'var1___1']
        )

        output_df = merge_with_prev_year(mock_df, 1, year_col='year_col')
        pd.testing.assert_frame_equal(expected_df, output_df)

        expected_df = pd.DataFrame(
            [
                ['harry', 2018, 'kensho', 1, np.nan, np.nan],
                ['harry', 2019, 'TOT', 3, np.nan, np.nan],
                ['hallas', 2017, 'harvard', 3, np.nan, np.nan],
                ['hallas', 2018, 'harvard', 5, np.nan, np.nan],
                ['hallas', 2019, 'harvard', 6, 'harvard', 3],
            ],
            columns=['player_url', 'year_col', 'team_id', 'var1', 'team_id___2', 'var1___2']
        )

        output_df = merge_with_prev_year(mock_df, 2, year_col='year_col')
        pd.testing.assert_frame_equal(expected_df, output_df)



def _calc_delta(df):
    df = df.dropna(subset=['weight_avg', 'diff'])
    return (df['diff'] * df['weight_avg']).sum() / df['weight_avg'].sum()


# TODO: add test
def calculate_aging_curve(df, stat_col, weight_col, age_col='age'):
    """Calculate an aging curve with the delta method."""
    prev_weight_col = '{}___1'.format(weight_col)
    prev_stat_col = '{}___1'.format(stat_col)
    df = df.copy()[[stat_col, weight_col, age_col, prev_stat_col, prev_weight_col]]
    df['diff'] = df[stat_col] - df[prev_stat_col]
    df['weight_avg'] = 2 / (1 / df[weight_col] + 1 / df[prev_weight_col])
    age_delta = df.groupby(age_col).apply(_calc_delta)
    return age_delta.fillna(0).cumsum()



from ..scraping import get_stats_for_year, get_stats_for_years
from .table_names import ADVANCED_STATS_TABLE_NAME
from .url_formats import ADVANCED_STATS_URL_FORMATTER


def get_advanced_stats_for_year(year):
    """Fetch advanced NBA stats for given year."""
    return get_stats_for_year(ADVANCED_STATS_URL_FORMATTER, year, ADVANCED_STATS_TABLE_NAME)


def get_advanced_stats(start_year, end_year):
    """Fetch advanced NBA stats for given year range."""
    return get_stats_for_years(
        ADVANCED_STATS_URL_FORMATTER, start_year, end_year, ADVANCED_STATS_TABLE_NAME
    )



import unittest

import requests
from bs4 import BeautifulSoup

from ..parsing import get_table_from_soup, get_table_soup


class TestAsExpected(unittest.TestCase):

    def test_scraping(self):
        # This page should not be updated with new data as of this writing
        # The only changes that would occur would be structural
        url = "https://www.basketball-reference.com/leagues/NBA_2019_advanced.html"
        r = requests.get(url)
        soup = BeautifulSoup(r.content)
        table_str = get_table_soup(soup, 'advanced_stats')
        table_data = get_table_from_soup(table_str)
        self.assertEqual((530, 29), table_data.shape)



import os
import unittest

import pandas as pd
from bs4 import BeautifulSoup

from ..parsing import get_table_from_soup, get_table_soup, list_tables

data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


class TestParsing(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        with open(os.path.join(data_path, 'mock_page.txt')) as f:
            content = f.read()
        cls.soup = BeautifulSoup(content, "lxml")
        with open(os.path.join(data_path, 'mock_table.txt')) as f:
            cls.expected_base_tag = BeautifulSoup(f.read(), "lxml").find('table')
        with open(os.path.join(data_path, 'mock_table_in_comments.txt')) as f:
            cls.expected_comment_tag = BeautifulSoup(f.read(), "lxml").find('table')

    def test_get_table_soup(self):
        # Test with regular table
        table_tag = get_table_soup(self.soup, 'confs_standings_E')
        self.assertEqual(self.expected_base_tag, table_tag)

        # Test with table in comments
        table_tag = get_table_soup(self.soup, 'misc_stats')
        self.assertEqual(self.expected_comment_tag, table_tag)

    def test_get_table_from_soup(self):
        # Test with normal table
        table_data = get_table_from_soup(self.expected_base_tag)
        expected_df = pd.read_csv(os.path.join(data_path, 'mock_data.csv'))
        pd.testing.assert_frame_equal(expected_df.drop('team_name_url', axis=1), table_data)

        # Test getting urls
        table_data = get_table_from_soup(self.expected_base_tag, get_url=True)
        pd.testing.assert_frame_equal(expected_df, table_data)

        # Test with headers on second row
        expected_df = pd.read_csv(
            os.path.join(data_path, 'mock_data_multiheader.csv'), dtype={'ranker': float}
        )
        expected_df['arena_name'] = expected_df['arena_name'].fillna('')
        table_data = get_table_from_soup(self.expected_comment_tag, include_tfoot=True)
        pd.testing.assert_frame_equal(expected_df, table_data)

        # Test without tfoot
        table_data = get_table_from_soup(self.expected_comment_tag)
        expected_df = expected_df.iloc[:-1]
        for col in ['wins', 'losses', 'ranker']:
            expected_df[col] = expected_df[col].astype(int)
        pd.testing.assert_frame_equal(expected_df, table_data)

    def test_list_tables(self):
        expected_result = [
            'confs_standings_E',
            'confs_standings_W',
            'divs_standings_E',
            'divs_standings_W',
            'all_playoffs',
            'team-stats-per_game',
            'opponent-stats-per_game',
            'team-stats-base',
            'opponent-stats-base',
            'team-stats-per_poss',
            'opponent-stats-per_poss',
            'misc_stats',
            'team_shooting',
            'opponent_shooting',
            'all_awards',
        ]
        found_tables = list_tables(self.soup)
        self.assertListEqual(expected_result, found_tables)



import unittest

import numpy as np
import pandas as pd

from ..stats import calculate_aging_curve


class TestStats(unittest.TestCase):

    def test_calculate_aging_curve(self):
        mock_df = pd.DataFrame(
            [
                [np.nan, 4., 2., 6., 19],
                [2., 1., 1., 3., 20],
                [2., 4., 2., 6., 20],
                [1., 4., 2., 6., 21],
                [3., np.nan, 2., 6., 21],
            ],
            columns=['col', 'col___1', 'mp', 'mp___1', 'age']
        )
        output_series = calculate_aging_curve(mock_df, 'col', 'mp')
        m1 = 2 / (1 / 1 + 1 / 3)
        m2 = 2 / (1 / 2 + 1 / 6)
        expected_series = pd.Series(
            [
                0.,  # fillna to 0
                (1. * m1 + -2 * m2) / (m1 + m2),
                -4,  # cumulative sum
            ],
            index=pd.Index([19, 20, 21], name='age')
        )
        pd.testing.assert_series_equal(expected_series, output_series)



from ..scraping import get_stats_for_year, get_stats_for_years
from .table_names import PASSING_TABLE_NAME, RUSHING_TABLE_NAME
from .url_formats import PASSING_URL_FORMATTER, RUSHING_URL_FORMATTER


def get_stats(stat, start_year, end_year):
    """Fetch NFL stats for given year range and specific stat types."""
    stat_types = {
        'passing': (PASSING_URL_FORMATTER, PASSING_TABLE_NAME),
        'rushing': (RUSHING_URL_FORMATTER, RUSHING_TABLE_NAME),
    }
    if stat not in stat_types:
        raise ValueError(
            'stat type {} is not supported currently. Only supported types are {}'.format(
                stat, stat_types.keys()
            )
        )
    url_formatter, table_name = stat_types[stat]
    return get_stats_for_years(url_formatter, start_year, end_year, table_name)



import numpy as np
import pandas as pd
from bs4 import BeautifulSoup


def _clean_soup(soup):
    """Clean up html to get rid of tricky comments."""
    return BeautifulSoup(str(soup).replace('<!--', '').replace('-->', ''), "lxml")


def _get_urls(row):
    """Get urls for all cells in table row."""
    all = row.findAll('th') + row.findAll('td')
    return [t.find('a')['href'] if t.find('a') else np.nan for t in all]


def list_tables(soup):
    """List tables present in BSoup object that we know how to fetch."""
    soup = _clean_soup(soup)
    return [t.attrs['id'] for t in soup.findAll('table') if 'id' in t.attrs]


def _convert_dtypes(df):
    """Convert dtypes of df, when possible."""
    for col in df.columns:
        try:
            df[col] = [int(i) if i != '' and i is not None else None for i in df[col]]
        except ValueError:
            try:
                df[col] = df[col].replace('', np.nan).astype(float)
            except ValueError:
                pass
    return df


def get_table_from_soup(table_soup, get_url=False, include_tfoot=False):
    """Extract table from html to pd.DataFrame."""
    th_soups = table_soup.find('thead').findAll('tr')
    columns = [t['data-stat'] for t in th_soups[-1].findAll('th')]
    row_soups = table_soup.findAll('tr', {'class': 'full_table'})
    if not row_soups:
        row_soups = table_soup.find('tbody').findAll('tr')
    if include_tfoot:
        row_soups.extend(table_soup.find('tfoot').findAll('tr'))
    rows = [[r.find('th').text] + [t.text for t in r.findAll('td')] for r in row_soups]
    df = pd.DataFrame(rows, columns=columns)
    if get_url:
        extra_rows = [_get_urls(r) for r in row_soups]
        extra_df = pd.DataFrame(extra_rows, columns=[c + '_url' for c in columns])
        non_empty_cols = extra_df.notnull().sum()[lambda x: x != 0].index
        df = pd.concat([df, extra_df[non_empty_cols]], axis=1)
    return _convert_dtypes(df)


def get_table_soup(soup, table_name):
    """Find html for table, even if in a comment."""
    soup = _clean_soup(soup)
    tables = soup.findAll('table', {"id": table_name})
    if tables:
        return tables[0]
    else:
        raise ValueError('table {} not found'.format(table_name))



from os import path
from setuptools import setup, find_packages

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='pysportsref',
      version='0.0.2',
      packages=find_packages(),
      description='Utilities for working with Sports Reference',
      install_requires=['bs4', 'numpy', 'pandas', 'requests', 'lxml', 'tqdm'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='MIT',
      url='https://github.com/hwchase17/pysportsref')



from functools import partial


def _get_id(df, player_col, year_col):
    """Aggregate player column and year column."""
    return df[player_col] + '___' + df[year_col].astype(str)


def get_player_totals(df, team_col='team_id', player_col='player_url', year_col='year'):
    """If a player is present multiple times for the same year, get their TOT stats."""
    missing_cols = {team_col, player_col, year_col}.difference(df.columns)
    if missing_cols:
        raise ValueError('Missing columns: {}'.format(missing_cols))
    _partialed_get_id = partial(_get_id, player_col=player_col, year_col=year_col)
    vals = set(_partialed_get_id(df[df[team_col] == 'TOT']))
    return df[~(_partialed_get_id(df).isin(vals) & (df[team_col] != 'TOT'))].reset_index(drop=True)


def merge_with_prev_year(df, year_delta, year_col='year', player_col='player_url'):
    """Merge player stats from previous year."""
    if year_delta < 1:
        raise ValueError('year_delta must be >= 1, got {}!'.format(year_delta))
    prev_year_df = df.copy()
    prev_year_df[year_col] += year_delta
    merged_df = df.merge(
        prev_year_df,
        how='left',
        on=[player_col, year_col],
        suffixes=('', '___{}'.format(year_delta))
    )
    return merged_df



import pandas as pd
import requests
from bs4 import BeautifulSoup
from tqdm import tqdm

from .parsing import get_table_from_soup, get_table_soup


def get_soup(url):
    """Get beautiful soup representation of a url."""
    r = requests.get(url)
    return BeautifulSoup(r.content, features="html.parser")


def get_stats_for_year(url_formatter, year, table_name):
    """Get html for page for given year, find and extract table, add year column."""
    url = url_formatter.format(year=year)
    soup = get_soup(url)
    table_tag = get_table_soup(soup, table_name)
    table_data = get_table_from_soup(table_tag, get_url=True)
    table_data['year'] = year
    return table_data


def get_stats_for_years(url_formatter, start_year, end_year, table_name):
    """Get table data from url for range of years."""
    all_dfs = []
    for year in tqdm(range(start_year, end_year + 1)):
        all_dfs.append(get_stats_for_year(url_formatter, year, table_name))
    return pd.concat(all_dfs, sort=True).reset_index(drop=True)
